from __future__ import annotations
from . import ScratchValue, ScratchVariable, ScratchList
from . import uid as get_uid
from enum import Enum, StrEnum
import json

class StackBlock():
    def __init__(self, op: str, ins: dict[str, BlockInput | list], fields: dict[str, list], top: bool = False, substack: SubStack | None = None, substack2: SubStack | None = None):
        self.opcode = op
        self.ins = ins
        self.fields = fields
        self.top = top
        self.uid = get_uid()
        self.substack = substack
        self.substack2 = substack2
        self.parentuid = None
        self.nextuid = None
        self.shadow = False
        self.mutation: dict | None = None
        if self.substack:
            self.ins["SUBSTACK"] = [2, self.substack.stack_blocks[0].uid]
            self.substack.stack_blocks[0].parentuid = self.uid
        if self.substack2:
            self.ins["SUBSTACK2"] = [2, self.substack2.stack_blocks[0].uid]
            self.substack2.stack_blocks[0].parentuid = self.uid
    def __iter__(self):
        yield ("opcode", self.opcode)
        yield ("next", self.nextuid)
        yield ("parent", self.parentuid)
        try:
            yield ("inputs", dict([(k, list(v)) for k, v in self.ins.items()]))
        except TypeError as e:
            print(self.ins.items())
            print(f"Caught type error in input assignment during block creation: {e}")
            yield ("inputs", {})
        yield ("fields", self.fields)
        yield ("shadow", self.shadow)
        yield ("topLevel", self.top)
        if self.mutation:
            yield ("mutation", self.mutation)

class OpBlock(StackBlock):
    def __init__(self, op, ins, fields, top = False, substack = None, substack2 = None):
        super().__init__(op, ins, fields, top, substack, substack2)

class SubStack():
    def __init__(self):
        self.stack_blocks: list[StackBlock] = []
        self.op_blocks: list[OpBlock] = []
    def push_stack(self, block: StackBlock):
        self.stack_blocks.append(block)
    def push_op(self, block: OpBlock):
        self.op_blocks.append(block)
        return block
    def __iter__(self):
        for i in range(len(self.stack_blocks)):
            x = self.stack_blocks[i]
            try:
                x.nextuid = self.stack_blocks[i+1].uid
            except IndexError:
                pass
            try:
                x.parentuid = self.stack_blocks[i-1].uid
            except IndexError:
                pass
            yield (x.uid, dict(x))
        for i in range(len(self.op_blocks)):
            x = self.op_blocks[i]
            yield (x.uid, dict(x))

class Script(SubStack):
    def __init__(self, start: StackBlock):
        super().__init__()
        self.stack_blocks.append(start)

class BlockInput():
    def __init__(self, data: ScratchValue | ScratchVariable | ScratchList | OpBlock, shadow: bool | int, default: ScratchValue | None = None):
        self.data = data
        self.shadow = shadow
        if default is None and not isinstance(self.data, OpBlock):
            self.shadow_default = ScratchValue("")
            if isinstance(self.data, OpBlock):
                if isinstance(self.data.val, int):
                    self.shadow_default = ScratchValue(0)
        else:
            self.shadow_default = default
    def __iter__(self):
        if isinstance(self.shadow, bool):
            shadow = 1 if self.shadow else 2
            if self.shadow_default and shadow == 2:
                shadow = 3
        else:
            shadow = self.shadow
        yield shadow
        if isinstance(self.data, ScratchValue):
            if isinstance(self.data.val, int):
                yield [4, self.data.val]
            else:
                yield [10, str(self.data)]
        elif isinstance(self.data, ScratchVariable):
            yield [12, self.data.name, self.data.uid]
        elif isinstance(self.data, ScratchList):
            yield [13, self.data.name, self.data.uid]
        elif isinstance(self.data, OpBlock):
            newblock = self.data
            yield newblock.uid
        if shadow == 3 and not isinstance(self.data, ScratchValue):
            if not isinstance(self.data, OpBlock) and isinstance(self.data.val, int):
                yield [4, self.shadow_default.val]
            else:
                yield [10, str(self.shadow_default)]

class ProcedureInputType(StrEnum):
    STRING_NUMBER = "argument_reporter_string_number"
    BOOLEAN = "argument_reporter_boolean"

class ProcedureInput():
    def __init__(self, name: str, intype = ProcedureInputType.STRING_NUMBER):
        self.uid = get_uid()
        self.name = name
        self.intype: ProcedureInputType = intype
    def as_op(self):
        return OpBlock(self.intype, {}, {"VALUE": [self.name, None]})

class Procedure():
    def __init__(self, name: str, inputs: list[ProcedureInput] = [], warp: bool = False):
        ins = {}
        prototype_stack = SubStack()
        proccode = [name]
        argids = []
        argnames = []
        argdef = []
        for x in inputs:
            asop = x.as_op()
            ins[x.uid] = [1, prototype_stack.push_op(asop).uid]
            argids.append(x.uid)
            argnames.append(x.name)
            if x.intype is ProcedureInputType.STRING_NUMBER:
                proccode.append("%s")
                argdef.append("")
            elif x.intype is ProcedureInputType.BOOLEAN:
                proccode.append("%b")
                argdef.append("false")
        proccode = " ".join(proccode)
        self.proccode = proccode
        prototype = StackBlock("procedures_prototype", ins=ins, fields={})
        prototype.shadow = True
        self.mutation = {"tagName": "mutation",
                              "children": [],
                              "proccode": self.proccode,
                              "argumentids": json.dumps(argids),
                              "argumentnames": json.dumps(argnames),
                              "argumentdefaults": json.dumps(argdef),
                              "warp": json.dumps(warp)}
        prototype.mutation = self.mutation
        prototype_stack.push_stack(prototype)

        definition = Script(StackBlock("procedures_definition", {"custom_block": [1, prototype.uid]}, {}, top=True))

        self.definition = definition
        self.prototype = prototype_stack
        self.arguments = inputs
    def call_proc(self, ins: dict[ProcedureInput, list]):
        block = StackBlock("procedures_call", dict([(k.uid, v) for k, v in ins.items()]), {})
        block.mutation = self.mutation
        return block



from random import randrange

def uid(length: int = 20):
    """
    Generates a UID, see: https://github.com/scratchfoundation/scratch-vm/blob/develop/src/util/uid.js

    Directly translated from the original Javascript code
    """
    SOUP = "!#%()*+,-./:;=?@[]^_`{|}~ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
    id = ""
    for i in range(length):
        id += SOUP[randrange(0, len(SOUP))]
    return id

class ScratchValue():
    def __init__(self, value: str | int | bool):
        self.val = value
    def __str__(self) -> str:
        return str(self.val)

# Base class which represents Scratch variables and lists
class _ScratchObject():
    def __init__(self, name: str, value: ScratchValue | list[ScratchValue] = ScratchValue(""), id: str | None = None):
        if id is None:
            self.uid: str = uid()
        else:
            self.uid: str = id
        self.val: ScratchValue | list[ScratchValue] = value
        self.name: str = name
    def __iter__(self):
        yield self.name
        yield self.val
    def target_def_form(self) -> tuple:
        return (self.uid, list(self))

class ScratchVariable(_ScratchObject):
    def __init__(self, name: str, value: ScratchValue = ScratchValue(""), is_cloud: bool = False, uid: str | None = None):
        super().__init__(name, value, uid)
        self.is_cloud: bool = is_cloud
    def __iter__(self):
        yield from super().__iter__()
        if self.is_cloud:
            yield True

class ScratchList(_ScratchObject):
    def __init__(self, name: str, value: list[ScratchValue] = [ScratchValue("")], id: str | None = None):
        super().__init__(name, value, id)
    def __getitem__(self, i: int):
        return self.val[i]

class Broadcast():
    def __init__(self, name: str, id: str | None = None):
        self.name: str = name
        if id is None:
            self.uid = uid()
        else:
            self.uid = id
    def target_def_form(self) -> tuple:
        return (self.uid, self.name)



import json
import zipfile
import pathlib
from . import blocks, targets, scratchtypes as st

class ProjectFile():
    def __init__(self, targets: list[targets.BaseTarget], extensions: list[str] = [], vmver: str = "0.2.0", ua: str = "ScratchPython"):
        self.targets = targets
        self.extensions = extensions
        self.vmver = vmver
        self.agent = ua
        self.platform = {"name": "ScratchPython sb3 editor", "url": "https://github.com/Belu-cat/ScratchPython/tree/main/src/ScratchPython/sb3"}
    def __iter__(self):
        yield ("targets", [dict(x) for x in self.targets])
        yield ("monitors", [])
        yield ("extensions", self.extensions)
        yield ("meta", {"semver": "3.0.0",
                        "vm": self.vmver,
                        "agent": self.agent,
                        # "platform": self.platform})
        })

def project_to_zip(projectfile: ProjectFile, outfilename: str):
    with zipfile.ZipFile(outfilename, 'w', zipfile.ZIP_DEFLATED) as zf:
        zf.writestr("project.json", json.dumps(dict(projectfile)))
        zf.writestr(targets.BLANKSVGHASH+".svg", targets.BLANKSVG)



# Code for sb3 "targets", see https://en.scratch-wiki.info/wiki/Scratch_File_Format#Targets

from . import uid, ScratchValue, ScratchList, ScratchVariable, Broadcast, blocks

BLANKSVG = """<svg xmlns="http://www.w3.org/2000/svg" width="1" height="1"/>"""
BLANKSVGHASH = "ad97b1aa2e5e1ca965c8221e21f09341"

# Base class from which all targets are derived
class BaseTarget():
    def __init__(self):
        self.variables: list[ScratchVariable] = []
        self.lists: list[ScratchList] = []
        self.blocks: list[blocks.SubStack] = []
        # self.comments = [] #TODO: implement Comment object
    def __iter__(self):
        yield ("variables", dict([x.target_def_form() for x in self.variables]))
        yield ("lists", dict([x.target_def_form() for x in self.lists]))
        b = {}
        for x in self.blocks:
            b |= dict(x)
        yield ("blocks", b)
        yield ("comments", {})
        yield ("currentCostume", 0)
        yield ("costumes", [{"assetId": BLANKSVGHASH, "name": "blank", "md5ext": BLANKSVGHASH+".svg", "dataFormat": "svg", "bitmapResolution": 1, "rotationCenterX":0,"rotationCenterY":0}])
        yield ("sounds", [])
        yield ("volume", 0)
    def push_substack(self, substack: blocks.SubStack):
        self.blocks.append(substack)
    def create_variable(self, name, value):
        v = ScratchVariable(name, value)
        self.variables.append(v)
        return v
    def create_list(self, name, value: list):
        l = ScratchList(name, value)
        self.lists.append(l)
        return l
    def push_proc(self, proc: blocks.Procedure):
        self.push_substack(proc.definition)
        self.push_substack(proc.prototype)

class Stage(BaseTarget):
    def __init__(self):
        super().__init__()
        self.broadcasts: list[Broadcast] = []
    def __iter__(self):
        yield ("isStage", True)
        yield ("name", "Stage")
        yield ("broadcasts", dict([(x.uid, x.name) for x in self.broadcasts]))
        yield from super().__iter__()
        yield ("tempo", 60)
        yield ("videoState", "off")
        yield ("videoTransparency", 50)
        yield ("textToSpeechLanguage", None)

class Sprite(BaseTarget):
    def __init__(self, name: str):
        super().__init__()
        self.name: str = name
    def __iter__(self):
        yield ("isStage", False)
        yield ("name", self.name)
        yield from super().__iter__()
        yield ("visible", False)
        yield ("x", 0)
        yield ("y", 0)
        yield ("size", 100)
        yield ("direction", 90)
        yield ("draggable", False)
        yield ("rotationStyle", "don't rotate")


import sb3gen.blocks as b
import sb3gen.targets as t
import sb3gen.project as p

# Create stage
stage = t.Stage()
# Create main script
script = b.Script(b.StackBlock("event_whenflagclicked", {}, {}, top=True))
# Create variable `var`
v = stage.create_variable("var", "value")
# Create list `list`
l = stage.create_list("list", ["example"])
# Create and push (1 + 1) block
inner = script.push_op(b.OpBlock("operator_add", {"NUM1": b.BlockInput(b.ScratchValue(1), True),"NUM2": b.BlockInput(b.ScratchValue(1), True)}, {}))
# Push `set variable [var] to (1 + 1)`
script.push_stack(b.StackBlock("data_setvariableto", {"VALUE": b.BlockInput(inner, False, default=b.ScratchValue(""))}, {"VARIABLE": [v.name, v.uid]}))
# Create substack
subs = b.SubStack()
# Create and push (1 + 1) block
inner2 = script.push_op(b.OpBlock("operator_add", {"NUM1": b.BlockInput(b.ScratchValue(1), True),"NUM2": b.BlockInput(b.ScratchValue(1), True)}, {}))
# inner = b.ScratchValue(2)
# Push `change variable [var] by (1+1)` to substack
subs.push_stack(b.StackBlock("data_changevariableby", {"VALUE": b.BlockInput(inner2, False)}, {"VARIABLE": [v.name, v.uid]}))
# Push the substack to the stage
stage.push_substack(subs)
# Push `forever {substack}` to the main script
script.push_stack(b.StackBlock("control_forever", {}, {}, substack=subs))
# Push the main script to the stage
stage.push_substack(script)

# Create procedure with an input named `test`
test = b.ProcedureInput("test")
proc = b.Procedure("example", [test])
# Create a variable named `variable`
v2 = stage.create_variable("variable", 0)
# Push `change variable [variable] by (1)` to procedure definition
proc.definition.push_stack(b.StackBlock("data_changevariableby", {"VALUE": b.BlockInput(b.ScratchValue(1), True)}, {"VARIABLE": [v2.name, v2.uid]}))
# Push `procedure example (variable)` to the procedure
proc.definition.push_stack(proc.call_proc({proc.arguments[0]: b.BlockInput(v2, False)}))
# Push this procedure to the stage
stage.push_proc(proc)

pf = p.ProjectFile([stage])
p.project_to_zip(pf, "example.sb3")
